// Skeleton Program code for the AQA A Level Paper 1 Summer 2020 examination
// this code should be used in conjunction with the Preliminary Material
// written by the AQA Programmer Team
// developed in the Delphi and Lazarus programming environments

program ALevel7517_2020_final;
{$APPTYPE CONSOLE}

uses
  SysUtils,
  Math;

var
  NextID : Integer = 1;

type
  Household = class
    protected
      ChanceEatOutPerDay : Real;
      XCoord, YCoord, ID : Integer;
    public
      constructor New(X : Integer; Y : Integer);
      function GetDetails() : String;
      function GetChanceEatOut() : Real;
      function GetX() : Integer;
      function GetY() : Integer;
  end;

type
  THouseholdArray = array of Household;

type
  Settlement = class
    protected
      StartNoOfHouseholds, XSize, YSize : Integer;
      Households : THouseholdArray;
      procedure CreateHouseholds();
    public
      constructor New();
      function GetNumberOfHouseholds() : Integer;
      function GetXSize() : Integer;
      function GetYSize() : Integer;
      procedure GetRandomLocation(var X : Integer; var Y : Integer);
      procedure AddHousehold();
      procedure DisplayHouseholds();
      function FindOutIfHouseholdEatsOut(HouseholdNo : Integer; var X : Integer; var Y : Integer) : Boolean;
  end;

type
  LargeSettlement = class(Settlement)
    public
      constructor New(ExtraXSize : Integer; ExtraYSize : Integer; ExtraHouseholds : Integer); overload;
  end;

type
  Outlet = class
    protected
      VisitsToday, XCoord, YCoord, Capacity, MaxCapacity : Integer;
      DailyCosts : Real;
    public
      constructor New(XCoordInput : Integer; YCoordInput : Integer; MaxCapacityBase : Integer);
      function GetCapacity() : Integer;
      function GetX() : Integer;
      function GetY() : Integer;
      procedure AlterDailyCost(Amount : Real);
      function AlterCapacity(Change : Integer) : Integer;
      procedure IncrementVisits();
      procedure NewDay();
      function CalculateDailyProfitLoss(AvgCostPerMeal : Real; AvgPricePerMeal : Real) : Real;
      function GetDetails() : String;
  end;

type
  TOutletArray = array of Outlet;
  TIntegerArray = array of Integer;

type
  Company = class
    protected
      Name, Category : String;
      Balance, ReputationScore, AvgCostPerMeal, AvgPricePerMeal, DailyCosts, FamilyOutletCost, FastFoodOutletCost, NamedChefOutletCost, FuelCostPerUnit, BaseCostOfDelivery : Real;
      Outlets: TOutletArray;
      FamilyFoodOutletCapacity, FastfoodOutletCapacity, NamedChefOutletCapacity : Integer;
    public
      constructor New(NameInput : String; CategoryInput : String; BalanceInput : Real; X : Integer; Y : Integer; FuelCostPerUnitInput : Real; BaseCostOfDeliveryInput : Single);
      function GetName() : String;
      function GetNumberofOutlets() : Integer;
      function GetReputationScore() : Real;
      procedure AlterDailyCosts(Change : Real);
      procedure AlterAvgCostPerMeal(Change : Real);
      procedure AlterFuelCostPerUnit(Change : Real);
      procedure AlterReputation(Change : Real);
      procedure NewDay();
      procedure AddVisitToNearestOutlet(X : Integer; Y : Integer);
      function GetDetails() : String;
      function ProcessDayEnd() : String;
      function CloseOutlet(ID : Integer) : Boolean;
      procedure ExpandOutlet(ID : Integer);
      procedure OpenOutlet(X : Integer; Y : Integer);
      function GetListOfOutlets() : TIntegerArray;
      function CalculateDeliveryCost() : Real;
    private
      function GetDistanceBetweenTwoOutlets(Outlet1 : Integer; Outlet2 : Integer) : Real;
  end;

type
  TCompanyArray = array of Company;

type
  Simulation = class
    protected
      SimulationSettlement : Settlement;
      NoOfCompanies : Integer;
      FuelCostPerUnit, BaseCostForDelivery : Real;
      Companies : TCompanyArray;
    public
      constructor New();
      procedure DisplayMenu();
      procedure ProcessDayEnd();
      procedure AddCompany();
      function GetIndexOfCompany(CompanyName : String) : Integer;
      procedure ModifyCompany(Index : Integer);
      procedure DisplayCompanies();
      procedure Run();
    private
      procedure DisplayCompaniesAtDayEnd();
      procedure ProcessAddHouseholdsEvent();
      procedure ProcessCostOfFuelChangeEvent();
      procedure ProcessReputationChangeEvent();
      procedure ProcessCostChangeEvent();
      procedure DisplayEventsAtDayEnd();
  end;

constructor Household.New(X : Integer; Y : Integer);
begin
  XCoord := X;
  YCoord := Y;
  ChanceEatOutPerDay := random();
  ID := NextID;
  inc(NextID);
end;

function Household.GetDetails() : String;
var
  Details : String;
begin
  Details := inttostr(ID) + '     Coordinates: (' + inttostr(XCoord) + ', ' + inttostr(YCoord) + ')     Eat out probability: ' + floattostr(ChanceEatOutPerDay);
  GetDetails := Details;
end;

function Household.GetChanceEatOut() : Real;
begin
  GetChanceEatOut := ChanceEatOutPerDay;
end;

function Household.GetX() : Integer;
begin
  GetX := XCoord;
end;

function Household.GetY() : Integer;
begin
  GetY := YCoord;
end;

constructor Settlement.New();
begin
  XSize := 1000;
  YSize := 1000;
  StartNoOfHouseholds := 250;
  SetLength(Households, 0);
  CreateHouseholds();
end;

function Settlement.GetNumberOfHouseholds() : Integer;
begin
  GetNumberOfHouseholds := length(Households);
end;

function Settlement.GetXSize() : Integer;
begin
  GetXSize := XSize;
end;

function Settlement.GetYSize() : Integer;
begin
  GetYSize := YSize;
end;

procedure Settlement.GetRandomLocation(var X : Integer; var Y : Integer);
begin
  X := round(random() * XSize);
  Y := round(random() * YSize);
end;

procedure Settlement.CreateHouseholds();
var
  Count : Integer;
begin
  for Count := 0 to Self.StartNoOfHouseholds - 1 do
    AddHousehold();
end;

procedure Settlement.AddHousehold();
var
  X, Y : Integer;
  Temp : Household;
begin
  SetLength(Households,length(Households) + 1);
  GetRandomLocation(X, Y);
  Temp := Household.New(X, Y);
  Households[length(Households) - 1] := Temp;
end;

procedure Settlement.DisplayHouseholds();
var
  H : Integer;
begin
  writeln;
  writeln('**********************************');
  writeln('*** Details of all households: ***');
  writeln('**********************************');
  writeln;
  for H := low(Self.Households) to high(Self.Households) do
    writeln(Self.Households[H].GetDetails());
  writeln;
end;

function Settlement.FindOutIfHouseholdEatsOut(HouseholdNo : Integer; var X : Integer; var Y : Integer) : Boolean;
var
  EatOutRNo : Real;
begin
  EatOutRNo := random();
  X := Households[HouseholdNo].GetX();
  Y := Households[HouseholdNo].GetY();
  if EatOutRNo < Households[HouseholdNo].GetChanceEatOut() then
    FindOutIfHouseholdEatsOut := True
  else
    FindOutIfHouseholdEatsOut := False;
end;

constructor LargeSettlement.New(ExtraXSize : Integer; ExtraYSize : Integer; ExtraHouseholds : Integer);
var
  Count : Integer;
begin
  Self.New();
  Self.XSize := Self.XSize + ExtraXSize;
  Self.YSize := Self.YSize + ExtraYSize;
  Self.StartNoOfHouseholds := Self.StartNoOfHouseholds + ExtraHouseholds;
  for Count := 1 to ExtraHouseholds do
    Self.AddHousehold();
end;

constructor Outlet.New(XCoordInput : Integer; YCoordInput : Integer; MaxCapacityBase : Integer);
begin
  Self.XCoord := XCoordInput;
  Self.YCoord := YCoordInput;
  Self.Capacity := trunc(MaxCapacityBase * 0.6);
  Self.MaxCapacity := MaxCapacityBase + trunc(random() * 50) - trunc(random() * 50);
  Self.DailyCosts := MaxCapacityBase * 0.2 + Capacity * 0.5 + 100;
  NewDay();
end;

function Outlet.GetCapacity() : Integer;
begin
  getCapacity := Self.Capacity;
end;

function Outlet.GetX() : Integer;
begin
  GetX := Self.XCoord;
end;

function Outlet.GetY() : Integer;
begin
  GetY := Self.YCoord;
end;

procedure Outlet.AlterDailyCost(Amount : Real);
begin
  Self.DailyCosts := Self.DailyCosts + Amount;
end;

function Outlet.AlterCapacity(Change : Integer) : Integer;
var
  OldCapacity : Integer;
begin
  OldCapacity := Self.Capacity;
  Self.Capacity := Self.Capacity + Change;
  if Self.Capacity > Self.MaxCapacity then
    begin
      Self.Capacity := Self.MaxCapacity;
      AlterCapacity := Self.MaxCapacity - OldCapacity;
    end
  else
    begin
      if Self.Capacity < 0 then
        Self.Capacity := 0;
      Self.DailyCosts := Self.MaxCapacity * 0.2 + Capacity * 0.5 + 100;
      AlterCapacity := Change;
    end;
end;

procedure Outlet.IncrementVisits();
begin
  Self.VisitsToday := Self.VisitsToday + 1;
end;

procedure Outlet.NewDay();
begin
  Self.VisitsToday := 0;
end;

function Outlet.CalculateDailyProfitLoss(AvgCostPerMeal : Real; AvgPricePerMeal : Real) : Real;
begin
  CalculateDailyProfitLoss := (AvgPriceperMeal - AvgCostPerMeal) * Self.VisitsToday - Self.DailyCosts;
end;

function Outlet.GetDetails() : String;
var
  Details : String;
begin
  Details := 'Coordinates: (' + inttostr(Self.XCoord) + ', ' + inttostr(Self.YCoord) + ')     Capacity: ' + inttostr(Self.Capacity) + '      Maximum Capacity: ';
  Details := Details + inttostr(Self.MaxCapacity) + '      Daily Costs: ' + floattostr(Self.DailyCosts) + '      Visits today: ' + inttostr(Self.VisitsToday);
  GetDetails := Details;
end;

constructor Company.New(NameInput : String; CategoryInput : String; BalanceInput : Real; X : Integer; Y : Integer; FuelCostPerUnitInput : Real; BaseCostOfDeliveryInput : Single);
begin
  Self.FamilyOutletCost := 1000;
  Self.FastFoodOutletCost := 2000;
  Self.NamedChefOutletCost:= 15000;
  Self.FamilyFoodOutletCapacity := 150;
  Self.FastFoodOutletCapacity := 200;
  Self.NamedChefOutletCapacity := 50;
  Self.Name := NameInput;
  Self.Category := CategoryInput;
  Self.Balance := BalanceInput;
  Self.FuelCostPerUnit := FuelCostPerUnitInput;
  Self.BaseCostOfDelivery := BaseCostOfDeliveryInput;
  Self.ReputationScore := 100;
  Self.DailyCosts := 100;
  if Self.Category = 'fast food' then
    begin
      Self.AvgCostPerMeal := 5;
      Self.AvgPricePerMeal := 10;
      Self.ReputationScore := Self.ReputationScore + random() * 10 - 8;
    end
  else if Self.Category = 'family' then
    begin
      Self.AvgCostPerMeal := 12;
      Self.AvgPricePerMeal := 14;
      Self.ReputationScore := Self.ReputationScore + random() * 30 - 5;
    end
  else
    begin
      Self.AvgCostPerMeal := 20;
      Self.AvgPricePerMeal := 40;
      Self.ReputationScore := Self.ReputationScore + random() * 50;
    end;
  OpenOutlet(X, Y);
end;

function Company.GetName() : String;
begin
  GetName := Self.Name;
end;

function Company.GetNumberofOutlets() : Integer;
begin
  GetNumberOfOutlets := length(Self.Outlets);
end;

function Company.GetReputationScore() : Real;
begin
  GetReputationScore := Self.ReputationScore;
end;

procedure Company.AlterDailyCosts(Change : Real);
begin
  Self.DailyCosts := Self.DailyCosts + Change;
end;

procedure Company.AlterAvgCostPerMeal(Change : Real);
begin
  Self.AvgCostPerMeal := Self.AvgCostPerMeal + Change;
end;

procedure Company.AlterFuelCostPerUnit(Change : Real);
begin
  Self.FuelCostPerUnit := Self.FuelCostPerUnit + Change;
end;

procedure Company.AlterReputation(Change : Real);
begin
  Self.ReputationScore := Self.ReputationScore + Change;
end;

procedure Company.NewDay();
var
  O : Integer;
begin
  for O := Low(Self.Outlets) to High(Self.Outlets) do
    Self.Outlets[O].NewDay();
end;

procedure Company.AddVisitToNearestOutlet(X : Integer; Y : Integer);
var
  NearestOutlet, Current : Integer;
  NearestOutletDistance, CurrentDistance : Real;
begin
  NearestOutlet := 0;
  NearestOutletDistance := sqrt(Sqr(Self.Outlets[0].GetX() - X) + Sqr(Outlets[0].GetY() - Y));
  for Current := 1 to length(Self.Outlets) - 1 do
    begin
      CurrentDistance := sqrt(sqr(Self.Outlets[Current].getX() - X) + sqr(Self.Outlets[Current].getY() - Y));
      if CurrentDistance < NearestOutletDistance then
        begin
          NearestOutletDistance := CurrentDistance;
          NearestOutlet := Current;
        end;
    end;
  Self.Outlets[NearestOutlet].IncrementVisits();
end;

function Company.GetDetails() : String;
var
  Details : String;
  Current : Integer;
begin
  Details := 'Name: ' + Self.Name + #13#10 + 'Type of business: ' + Self.Category + #13#10;
  Details := Details + 'Current balance: ' + floattostr(Self.Balance) + #13#10 + 'Average cost per meal: ' + floattostr(Self.AvgCostPerMeal) + #13#10;
  Details := Details + 'Average price per meal: ' + floattostr(self.AvgPricePerMeal) + #13#10 + 'Daily costs: ' + floattostr(self.DailyCosts) + #13#10;
  Details := Details + 'Delivery costs: ' + floattostr(self.CalculateDeliveryCost()) + #13#10 + 'Reputation: ' + floattostr(Self.ReputationScore) + #13#10;
  Details := Details + 'Number of outlets: ' + inttostr(length(self.Outlets)) + #13#10 + 'Outlets' + #13#10;
  for Current := 1 to length(Self.Outlets) do
    Details := Details + inttostr(Current) + '. ' + Self.Outlets[Current - 1].GetDetails() + #13#10;
  GetDetails := Details;
end;

function Company.ProcessDayEnd() : String;
var
  Details : String;
  ProfitLossFromOutlets, ProfitLossFromThisOutlet, DeliveryCosts : Real;
  Current : Integer;
begin
  ProfitLossFromOutlets := 0;
  ProfitLossFromThisOutlet := 0;
  if length(Self.Outlets) > 1 then
    DeliveryCosts := Self.BaseCostOfDelivery + CalculateDeliveryCost()
  else
    DeliveryCosts := Self.BaseCostOfDelivery;
  Details := 'Daily costs for company: ' + floattostr(Self.DailyCosts) + #13#10 + 'Cost for delivering produce to outlets: ' + floattostr(DeliveryCosts) + #13#10;
  for Current := 0 to length(Self.Outlets) - 1 do
    begin
      ProfitLossFromthisOutlet := Self.Outlets[Current].CalculateDailyProfitLoss(Self.AvgCostPerMeal,Self.AvgPricePerMeal);
      Details := Details + 'Outlet ' + inttostr(Current + 1) + ' profit/loss: ' + floattostr(ProfitLossfromThisOutlet) + #13#10;
      ProfitLossFromOutlets := ProfitLossFromOutlets + ProfitLossFromThisOutlet;
    end;
  Details := Details + 'Previous balance for company: ' + floattostr(self.Balance) + #13#10;
  Self.Balance := Self.Balance + ProfitLossFromOutlets - Self.DailyCosts - DeliveryCosts;
  Details := Details + 'New balance for company: ' + floattostr(Self.Balance);
  ProcessDayEnd := Details;
end;

function Company.CloseOutlet(ID : Integer) : Boolean;
var
  CloseCompany : Boolean;
  Index : Integer;
begin
  CloseCompany := False;
  for Index := ID to length(Self.Outlets) - 2 do
    Self.Outlets[Index] := Self.Outlets[Index + 1];
  SetLength(Self.Outlets, length(Self.Outlets) - 1);
  if length(Self.Outlets) = 0 then
    CloseCompany := True;
  CloseOutlet := CloseCompany;
end;

procedure Company.ExpandOutlet(ID : Integer);
var
  Change, Result : Integer;
  Input : String;
begin
  write('Enter amount you would like to expand the capacity by: ');
  readln(Input);
  Change := strtoint(Input);
  Result := Self.Outlets[ID].AlterCapacity(Change);
  if Result = Change then
    writeln('Capacity adjusted.')
  else
    writeln('Only some of that capacity added, outlet now at maximum capacity.');
end;

procedure Company.OpenOutlet(X : Integer; Y : Integer);
var
  Capacity : Integer;
  NewOutlet : Outlet;
begin
  if Self.Category = 'fast food' then
    begin
      Self.Balance := Self.Balance - Self.FastFoodOutletCost;
      Capacity := Self.FastfoodOutletCapacity;
    end
  else if Self.Category = 'family' then
    begin
      Self.Balance := Self.Balance - Self.FamilyOutletCost;
      Capacity := Self.FamilyFoodOutletCapacity;
    end
  else
    begin
      Self.Balance := Self.Balance - Self.NamedChefOutletCost;
      Capacity := Self.NamedChefOutletCapacity;
    end;
  NewOutlet := Outlet.New(X, Y, Capacity);
  SetLength(Self.Outlets, length(Self.Outlets) + 1);
  Self.Outlets[length(Self.Outlets) - 1] := NewOutlet;
end;

function Company.GetListOfOutlets() : TIntegerArray;
var
  Temp : TIntegerArray;
  Current : Integer;
begin
  SetLength(Temp, length(Self.Outlets));
  for Current := 0 to length(Self.Outlets) - 1 do
    begin
      Temp[Current] := Current;
    end;
  GetListOfOutlets := Temp;
end;

function Company.CalculateDeliveryCost() : Real;
var
  ListOfOutlets : TIntegerArray;
  TotalDistance, TotalCost : Real;
  Current : Integer;
begin
  ListOfOutlets := GetListOfOutlets();
  TotalDistance := 0;
  for Current := 0 to length(ListOfOutlets) - 2 do
    TotalDistance := TotalDistance + GetDistanceBetweenTwoOutlets(ListOfOutlets[Current], ListOfOutlets[Current + 1]);
  TotalCost := TotalDistance * Self.FuelCostPerUnit;
  CalculateDeliveryCost := TotalCost;
end;

function Company.GetDistanceBetweenTwoOutlets(Outlet1 : Integer; Outlet2 : Integer) : Real;
begin
  GetDistanceBetweenTwoOutlets := sqrt(sqr(Self.Outlets[Outlet1].GetX() - Self.Outlets[Outlet2].GetX()) + sqr(Self.Outlets[Outlet1].GetY() - Self.Outlets[Outlet2].GetY()));
end;

constructor Simulation.New();
var
  Choice : String;
  ExtraX, ExtraY, ExtraHouseholds, Count : Integer;
  Company1, Company2, Company3 : Company;
begin
  Self.FuelCostPerUnit := 0.0098;
  Self.BaseCostForDelivery := 100;
  write('Enter L for a large settlement, anything else for a normal size settlement: ');
  readln(Choice);
  if Choice = 'L' then
    begin
      write('Enter additional amount to add to X size of settlement: ');
      readln(ExtraX);
      write('Enter additional amount to add to Y size of settlement: ');
      readln(ExtraY);
      write('Enter additional number of households to add to settlement: ');
      readln(ExtraHouseholds);
      Self.SimulationSettlement := LargeSettlement.New(ExtraX, ExtraY, ExtraHouseholds);
    end
  else
    Self.SimulationSettlement := Settlement.New();
  write('Enter D for default companies, anything else to add your own start companies: ');
  readln(Choice);
  if Choice = 'D' then
    begin
      Self.NoOfCompanies := 3;
      SetLength(Self.Companies, Self.NoOfCompanies);
      Company1 := Company.New('AQA Burgers', 'fast food', 100000, 200, 203, Self.FuelCostPerUnit, Self.BaseCostForDelivery);
      Self.Companies[0] := Company1;
      Self.Companies[0].OpenOutlet(300, 987);
      Self.Companies[0].OpenOutlet(500, 500);
      Self.Companies[0].OpenOutlet(305, 303);
      Self.Companies[0].OpenOutlet(874, 456);
      Self.Companies[0].OpenOutlet(23, 408);
      Self.Companies[0].OpenOutlet(412, 318);
      Company2 := Company.New('Ben Thor Cuisine', 'named chef', 100400, 390, 800, Self.FuelCostPerUnit, Self.BaseCostForDelivery);
      Self.Companies[1] := Company2;
      Company3 := Company.New('Paltry Poultry', 'fast food', 25000, 800, 390, Self.FuelCostPerUnit, Self.BaseCostForDelivery);
      Self.Companies[2] := Company3;
      Self.Companies[2].OpenOutlet(400, 390);
      Self.Companies[2].OpenOutlet(820, 370);
      Self.Companies[2].OpenOutlet(800, 600);
    end
  else
    begin
      write('Enter number of companies that exist at the start of simulation: ');
      readln(Self.NoOfCompanies);
      for Count := 1 to Self.NoOfCompanies do
        AddCompany();
    end;
end;

procedure Simulation.DisplayMenu();
begin
  writeln;
  writeln('*********************************');
  writeln('**********    MENU     **********');
  writeln('*********************************');
  writeln('1. Display details of households');
  writeln('2. Display details of companies');
  writeln('3. Modify company');
  writeln('4. Add new company');
  writeln('6. Advance to next day');
  writeln('Q. Quit');
  writeln;
  write('Enter your choice: ');
end;

procedure Simulation.ProcessDayEnd();
var
  TotalReputation : Real;
  Reputations : array of Real;
  CompanyRNo, Counter, Current, LoopMax, NoOfReps, C, X, Y : Integer;
begin
  TotalReputation := 0;
  NoOfReps := 0;
  for C := low(Self.Companies) to high(Self.Companies) do
    begin
      Self.Companies[C].NewDay();
      TotalReputation := TotalReputation + Self.Companies[C].GetReputationScore();
      inc(NoOfReps);
      Setlength(Reputations, NoOfReps);
      Reputations[NoOfReps - 1] := TotalReputation;
    end;
  LoopMax := Self.SimulationSettlement.GetNumberOfHouseholds - 1;
  for Counter := 0 to LoopMax do
    begin
      if Self.SimulationSettlement.FindOutIfHouseholdEatsOut(Counter, X, Y) then
        begin
          CompanyRNo := trunc(random() * (trunc(TotalReputation) + 1));
          Current := 0;
          while Current < length(Reputations) do
            begin
              if CompanyRNo < Reputations[Current] then
                begin
                  Self.Companies[Current].AddVisitToNearestOutlet(X, Y);
                  break;
                end;
              inc(Current);
            end;
        end;
    end;
  DisplayCompaniesAtDayEnd();
  DisplayEventsAtDayEnd();
end;

procedure Simulation.AddCompany();
var
  Balance, X, Y : Integer;
  CompanyName, TypeOfCompany : String;
  NewCompany : Company;
begin
  write('Enter a name for the company: ');
  readln(CompanyName);
  write('Enter the starting balance for the company: ');
  readln(Balance);
  repeat
    write('Enter 1 for a fast food company, 2 for a family company or 3 for a named chef company: ');
    readln(TypeOfCompany);
  until ((TypeOfCompany = '1') or (TypeOfCompany = '2') or (TypeOfCompany = '3'));
  if TypeOfCompany = '1' then
    TypeOfCompany := 'fast food'
  else if TypeOfCompany = '2' then
    TypeOfCompany := 'family'
  else
    TypeOfCompany := 'named chef';
  Self.SimulationSettlement.GetRandomLocation(X, Y);
  NewCompany := Company.New(CompanyName, TypeOfCompany, Balance, X, Y, FuelCostPerUnit, Self.BaseCostForDelivery);
  SetLength(Self.Companies, length(Self.Companies) + 1);
  Self.Companies[High(Self.Companies)] := NewCompany;
end;

function Simulation.GetIndexOfCompany(CompanyName : String) : Integer;
var
  Current, Index : Integer;
begin
  Index := -1;
  for Current := 0 to length(Self.Companies) - 1 do
    if LowerCase(Self.Companies[Current].GetName()) = LowerCase(CompanyName) then
      begin
        Index := Current;
        break;
      end;
  GetIndexOfCompany := Index;
end;

procedure Simulation.ModifyCompany(Index : Integer);
var
  Choice : String;
  OutletIndex, CompIndex, X, Y : Integer;
  CloseCompany : Boolean;
begin
  writeln;
  writeln('*********************************');
  writeln('*******  MODIFY COMPANY   *******');
  writeln('*********************************');
  writeln('1. Open new outlet');
  writeln('2. Close outlet');
  writeln('3. Expand outlet');
  writeln;
  write('Enter your choice: ');
  readln(Choice);
  writeln;
  if (Choice = '2') or (Choice = '3') then
    begin
      write('Enter ID of outlet: ');
      readln(OutletIndex);
      if (OutletIndex > 0) and (OutletIndex <= Self.Companies[Index].GetNumberofOutlets()) then
        if Choice = '2' then
          begin
            CloseCompany := Self.Companies[Index].CloseOutlet(OutletIndex - 1);
            if CloseCompany = True then
              begin
                writeln('That company has now closed down as it has no outlets.');
                for CompIndex := Index to length(Self.Companies) - 2 do
                  Self.Companies[CompIndex] := Self.Companies[CompIndex + 1];
                SetLength(Self.Companies, length(Self.Companies) - 1);
              end;
          end
        else
          Self.Companies[Index].ExpandOutlet(OutletIndex - 1)
      else
        writeln('Invalid outlet ID.');
     end
  else if Choice = '1' then
  begin
    write('Enter X coordinate for new outlet: ');
    readln(X);
    write('Enter Y coordinate for new outlet: ');
    readln(Y);
    if (X >= 0) and (X < Self.SimulationSettlement.GetXSize()) and (Y >= 0) and (Y < Self.SimulationSettlement.GetYSize()) then
      Self.Companies[Index].OpenOutlet(X, Y)
    else
      writeln('Invalid coordinates');
  end;
  writeln;
end;

procedure Simulation.DisplayCompanies();
var
  C : Integer;
begin
  writeln;
  writeln('*********************************');
  writeln('*** Details of all companies: ***');
  writeln('*********************************');
  writeln;
  for C := low(Self.Companies) to high(Self.Companies) do
    writeln(Self.Companies[C].GetDetails() + #13#10);
  writeln;
end;

procedure Simulation.Run();
var
  CompanyName : String;
  Choice : Char;
  Index : Integer;
begin
  Choice := ' ';
  while Choice <> 'Q' do
    begin
      DisplayMenu();
      readln(Choice);
      case (Choice) of
        '1': Self.SimulationSettlement.DisplayHouseholds();
        '2': DisplayCompanies();
        '3':
          begin
            repeat
              write('Enter company name: ');
              readln(CompanyName);
              Index := GetIndexOfCompany(CompanyName);
            until Index <> - 1;
            ModifyCompany(Index);
          end;
        '4': AddCompany();
        '6': ProcessDayEnd();
        'Q':
          begin
            writeln('Simulation finished, press Enter to close.');
            readln;
          end;
      end;
    end;
end;

procedure Simulation.DisplayCompaniesAtDayEnd();
var
  Details : String;
  Index : Integer;
begin
  writeln;
  writeln('**********************');
  writeln('***** Companies: *****');
  writeln('**********************');
  writeln;
  for Index := low(Self.Companies) to high(Self.Companies) do
    begin
      writeln(Self.Companies[Index].GetName());
      writeln;
      Details := Self.Companies[Index].ProcessDayEnd();
      writeln(Details);
      writeln;
    end;
end;

procedure Simulation.ProcessAddHouseholdsEvent();
var
  NoOfNewHouseholds : Integer;
  Count : Integer;
begin
  NoOfNewHouseholds := trunc(random() * 4) + 1;
  for Count := 1 to NoOfNewHouseholds do
    Self.SimulationSettlement.AddHousehold();
  writeln(inttostr(NoOfNewHouseholds) + ' new households have been added to the settlement.');
end;

procedure Simulation.ProcessCostOfFuelChangeEvent();
var
  FuelCostChange : Real;
  UpOrDown, CompanyNo : Integer;
begin
  FuelCostChange := (trunc(random() * 9) + 1) / 10;
  UpOrDown := trunc(random() * 2);
  CompanyNo := trunc(Random() * length(Self.Companies));
  if UpOrDown  = 0 then
    writeln('The cost of fuel has gone up by ' + floattostr(FuelCostChange) + ' for ' + Self.Companies[CompanyNo].GetName())
  else
    begin
      writeln('The cost of fuel has gone down by ' + floattostr(FuelCostChange) + ' for ' + Self.Companies[CompanyNo].GetName());
      FuelCostChange := FuelCostChange * -1;
    end;
  Self.Companies[CompanyNo].AlterFuelCostPerUnit(FuelCostChange);
end;

procedure Simulation.ProcessReputationChangeEvent();
var
  ReputationChange : Real;
  UpOrDown, CompanyNo : Integer;
begin
  ReputationChange := (trunc(random() * 9) + 1) / 10;
  UpOrDown := trunc(random() * 2);
  CompanyNo := trunc(Random() * length(Self.Companies));
  if UpOrDown  = 0 then
    writeln('The reputation of ' + Self.Companies[CompanyNo].GetName() + 'has gone up by ' + floattostr(ReputationChange))
  else
    begin
      writeln('The reputation of ' + Self.Companies[CompanyNo].GetName() + 'has gone down by ' + floattostr(ReputationChange));
      ReputationChange := ReputationChange * -1;
    end;
  Self.Companies[CompanyNo].AlterReputation(ReputationChange);
end;

procedure Simulation.ProcessCostChangeEvent();
var
  CostToChange, UpOrDown, CompanyNo : Integer;
  AmountOfChange : Real;
begin
  CostToChange := trunc(random() * 2);
  UpOrDown := trunc(random() * 2);
  CompanyNo := trunc(random() * length(Self.Companies));
  if CostToChange = 0 then
    begin
      AmountOfChange := (trunc(random() * 19) + 1) / 10;
      if UpOrDown = 0 then
        writeln('The daily costs for ' + Self.Companies[CompanyNo].GetName() + ' have gone up by ' + floattostr(AmountOfChange))
      else
        begin
          writeln('The daily costs for ' + Self.Companies[CompanyNo].GetName() + ' have gone down by ' + floattostr(AmountOfChange));
          AmountOfChange := AmountOfChange * -1;
        end;
      Self.Companies[CompanyNo].AlterDailyCosts(AmountOfChange);
    end
  else
    begin
      AmountOfChange := trunc((random() * 9) + 1) / 10;
      if UpOrDown = 0 then
        writeln('The average cost of a meal for ' + Self.Companies[CompanyNo].GetName() + ' have gone up by ' + floattostr(AmountOfChange))
      else
        begin
          writeln('The average cost of a meal for ' + Self.Companies[CompanyNo].GetName() + ' have gone down by ' + floattostr(AmountOfChange));
          AmountOfChange := AmountOfChange * -1;
        end;
      Self.Companies[CompanyNo].AlterAvgCostPerMeal(AmountOfChange);
    end;
end;

procedure Simulation.DisplayEventsAtDayEnd();
var
  EventRanNo : Real;
begin
  writeln;
  writeln('***********************');
  writeln('*****   Events:   *****');
  writeln('***********************');
  writeln;
  EventRanNo := random();
  if EventRanNo < 0.25 then
    begin
      EventRanNo := random();
      if EventRanNo < 0.25 then
        ProcessAddHouseholdsEvent();
      EventRanNo := random();
      if EventRanNo < 0.5 then
        ProcessCostOfFuelChangeEvent();
      EventRanNo := random();
      if EventRanNo < 0.5 then
        ProcessReputationChangeEvent();
      EventRanNo := random();
      if EventRanNo >= 0.5 then
        ProcessCostChangeEvent();
    end
    else
      writeln('No events.');
end;

var
  ThisSim : Simulation;

begin
  randomize;
  ThisSim := Simulation.New();
  ThisSim.Run();
end.